/*	
   joy2key
		This program polls the joystick device /dev/js0 and dispatches
		X events to a specific window based on joystick status.
		It is most useful for adding joystick support to X games that 
		don't nativly support the joystick, using joy2key to send
		XKeyPress messages in place of actually typing movement keys.
		This code is under the GNU Public License (see COPYING), 
		you can do whatever you want to it, and if you do anything really
		neat, I'd like to see it!

		- Peter Amstutz (amstpi@freenet.tlh.fl.us)

		special thanks to Zhivago of EFNet's #c for helping me out a LOT with
			the XSendMessage() code...

		The latest version of joy2key can be found in 
			ftp://sunsite.unc.edu/pub/Linux/X11/xutils
		or on joy2key's (rather spartan) website,
			http://www.freenet.tlh.fl.us/~amstpi/joy2key.html

		Revision History
		----------------
		1.0 (12 Aug 97) - First version.
		1.1 (04 Dec 97) - Thanks to Craig Lamparter (craiger@hemna.rose.hp.com), 
		                  joy2key now has support for four-buttoned 
				  joysticks/gamepads!  Also some bugfixes.
	        1.2 (28 Apr 98) - Pierre Juhen (pierre.juhen@wanadoo.fr) fixed a dumb 
		                  little bug and added the autorepeat feature!
*/

/* The standard joystick driver (current ver 0.8.0) only supports two buttons, 
but if you apply the patch joystick.patch (in this archive) to the joystick 
driver you can use four.  Then you can define this */
/* #define USE_FOUR_BUTTONS */

#define JOY2KEY_VERSION "1.2"

/* Threshhold values will vary quite a bit from computer to computer! */
/* The following are my threshhold values for my Cyrix 6x86/166 
   with a CH Gamepad. */
#define DEFAULT_X_THRESHOLD_HIGH 	700
#define DEFAULT_X_THRESHOLD_LOW 	300
#define DEFAULT_Y_THRESHOLD_HIGH 	700
#define DEFAULT_Y_THRESHOLD_LOW 	300
/* These are the threshholds that Pierre seems to like.  
   He probably has a faster computer than I */
/* 
#define DEFAULT_X_THRESHOLD_HIGH       2400
#define DEFAULT_X_THRESHOLD_LOW        1200
#define DEFAULT_Y_THRESHOLD_HIGH       2000
#define DEFAULT_Y_THRESHOLD_LOW        1000
*/

#define DEFAULT_UP_KEYSYM 	        XK_Up
#define DEFAULT_DOWN_KEYSYM 		XK_Down
#define DEFAULT_LEFT_KEYSYM 		XK_Left
#define DEFAULT_RIGHT_KEYSYM 		XK_Right
#define DEFAULT_BUTTON1_KEYSYM		XK_z
#define DEFAULT_BUTTON2_KEYSYM		XK_a
#ifdef USE_FOUR_BUTTONS
#define DEFAULT_BUTTON3_KEYSYM		XK_s
#define DEFAULT_BUTTON4_KEYSYM		XK_x
#endif
#define DEFAULT_FREQUENCY	       100
#define DEFAULT_AUTOREPEAT	       5
#define DEFAULT_DEVICE		       "/dev/js0"
#define TRUE                           1
#define FALSE                          0


#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <linux/joystick.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <X11/keysym.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>

void cleanup(int s);

Display *thedisp;

int main(int argc, char **argv)
{
    int thewindow=0, thescreen=0;
    FILE *foo;
    char string[255], winname[255], *device=DEFAULT_DEVICE;
    XEvent event;
    struct JS_DATA_TYPE js0;
    int i, jsfd,
	oldx, oldy, oldb,
	b1=0, b2=0, b3=0, b4=0, oldb1, oldb2, oldb3, oldb4,
	arup=DEFAULT_UP_KEYSYM, 
	ardown=DEFAULT_DOWN_KEYSYM, 
	arleft=DEFAULT_LEFT_KEYSYM, 
	arright=DEFAULT_RIGHT_KEYSYM, 
	but1=DEFAULT_BUTTON1_KEYSYM, 
	but2=DEFAULT_BUTTON2_KEYSYM,
#ifdef USE_FOUR_BUTTONS
	but3=DEFAULT_BUTTON3_KEYSYM, 
	but4=DEFAULT_BUTTON4_KEYSYM,
#endif
	xthreshhi=DEFAULT_X_THRESHOLD_HIGH,
	xthreshlow=DEFAULT_X_THRESHOLD_LOW,
	ythreshhi=DEFAULT_Y_THRESHOLD_HIGH,
	ythreshlow=DEFAULT_Y_THRESHOLD_LOW,
	autorepeatflag=FALSE,
	timeindex=1,
	elapsed=FALSE,
	justafterelapsed=FALSE;
    struct timeval  timeval[2], *lasttime=timeval, *presenttime=timeval+1;
    struct timezone timezone;
    long autorepeattime=DEFAULT_AUTOREPEAT, 
	waittime=DEFAULT_FREQUENCY;

    puts("joy2key - reads joystick status and dispatches X keyboard events");
    puts("By Peter Amstutz (amstpi@freenet.tlh.fl.us)");
    puts("This is free software under the GNU Public License");
    puts("              (see COPYING in the joy2key archive)");
    puts("You are welcome to use/modify this code, and please e-mail me");
    puts("if anything cool comes of it!");
    printf("Version: %s   Binary built on %s at %s\n\n", 
	   JOY2KEY_VERSION, __DATE__, __TIME__);

    thedisp=XOpenDisplay(NULL);
    if(thedisp==NULL) 
    {
	printf("Error opening X Display");
	return 0;
    }
    thescreen=DefaultScreen(thedisp);

    if(argc<2) 
    {
	puts("Usage: joy2key \"Window Name\"");
	printf("\n\t[ -arrows [%s] [%s] [%s] [%s] ]", 
	     XKeysymToString(DEFAULT_UP_KEYSYM),
	     XKeysymToString(DEFAULT_DOWN_KEYSYM),
	     XKeysymToString(DEFAULT_LEFT_KEYSYM),
	     XKeysymToString(DEFAULT_RIGHT_KEYSYM));
	printf("\n\t[ -buttons12 [%s] [%s] ]",
	     XKeysymToString(DEFAULT_BUTTON1_KEYSYM),
	     XKeysymToString(DEFAULT_BUTTON2_KEYSYM));
#ifdef USE_FOUR_BUTTONS
	printf("\n\t[ -buttons34 [%s] [%s] ]",
	     XKeysymToString(DEFAULT_BUTTON3_KEYSYM),
	     XKeysymToString(DEFAULT_BUTTON4_KEYSYM));
#endif
	printf("\n\t[ -xthresh [hi[%i]] [low[%i]] ]", DEFAULT_X_THRESHOLD_HIGH, DEFAULT_X_THRESHOLD_LOW);
	printf("\n\t[ -ythresh [hi[%i]] [low[%i]] ]", DEFAULT_Y_THRESHOLD_HIGH, DEFAULT_Y_THRESHOLD_LOW);
	printf("\n\t[ -freq [%i]] ]", DEFAULT_FREQUENCY);
	printf("\n\t[ -dev [%s]] ]", DEFAULT_DEVICE);
	printf("\n\t[ -autorepeat [freq[%i]] ]\n", DEFAULT_AUTOREPEAT);
	return 0;
    }

    sprintf(winname, "xwininfo -name \"%s\"", argv[1]);

    for(i=2; i<argc; i++)
    {
	if(!strcmp(argv[i], "-arrows"))
	{
	    if(i+5>argc) 
	    {
		puts("Not enough arguments to -arrows");
		return 0;
	    }
	    if((arup=XStringToKeysym(argv[++i]))==NoSymbol) 
		printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
	    if((ardown=XStringToKeysym(argv[++i]))==NoSymbol) 
		printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
	    if((arleft=XStringToKeysym(argv[++i]))==NoSymbol) 
		printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
	    if((arright=XStringToKeysym(argv[++i]))==NoSymbol) 
		printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
	    continue;
	}
	if(!strcmp(argv[i], "-buttons12"))
	{
	    if(i+3>argc) 
	    {
		puts("Not enough arguments to -buttons12");
		return 0;
	    }
	    if((but1=XStringToKeysym(argv[++i]))==NoSymbol) 
	    {
		printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
		but1=DEFAULT_BUTTON1_KEYSYM;
	    }
	    if((but2=XStringToKeysym(argv[++i]))==NoSymbol) 
	    {
		printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
		but2=DEFAULT_BUTTON2_KEYSYM;
	    }
	    continue;
	}
#ifdef USE_FOUR_BUTTONS
	if(!strcmp(argv[i], "-buttons34"))
	{
	    if(i+3>argc) 
	    {
		puts("Not enough arguments to -buttons34");
		return;
	    }
	    if((but3=XStringToKeysym(argv[++i]))==NoSymbol) 
	    {
		printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
		but3=DEFAULT_BUTTON3_KEYSYM;
	    }
	    if((but4=XStringToKeysym(argv[++i]))==NoSymbol) 
	    {
		printf("Can't find %s, check include/X11/keysymdef.h, using default\n", argv[i]);
		but4=DEFAULT_BUTTON4_KEYSYM;
	    }
	    continue;
	}
#endif
	if(!strcmp(argv[i], "-xthresh"))
	{
	    if(i+3>argc) 
	    {
		puts("Not enough arguments to -xthresh");
		return 0;
	    }
	    xthreshhi=atoi(argv[++i]);
	    xthreshlow=atoi(argv[++i]);
	    continue;
	}
	if(!strcmp(argv[i], "-ythresh"))
	{
	    if(i+3>argc) 
	    {
		puts("Not enough arguments to -ythresh");
		return 0;
	    }
	    ythreshhi=atoi(argv[++i]);
	    ythreshlow=atoi(argv[++i]);
	    continue;
	}
	if(!strcmp(argv[i], "-freq"))
	{
	    if(i+2>argc) 
	    {
		puts("Not enough arguments to -freq");
		return 0;
	    }
	    if((waittime=atoi(argv[++i]))==0) 
	    { 
		puts("Invalid polling frequency, using default");
		waittime=DEFAULT_FREQUENCY;
	    }
	    continue;
	}
	if(!strcmp(argv[i], "-autorepeat"))
	{ 			
	    if(i+2>argc || (argv[i+1][0]=='-')) 
	    {
		autorepeattime=DEFAULT_AUTOREPEAT;
		autorepeatflag=TRUE;
		i++;
		continue;
	    }
	    if((autorepeattime=atoi(argv[++i]))==0) 
	    { 
		puts("Invalid autorepeat frequency, using default");
		autorepeattime=DEFAULT_AUTOREPEAT;
	    }
	    if(autorepeattime>50) 
	    {
		autorepeattime=50;
		printf("Set to maximum 50 times / second.\n");
	    }
	    autorepeatflag=TRUE;
	    continue;
	}
	if(!strcmp(argv[i], "-dev"))
	{
	    if(i+2>argc) 
	    {
		puts("Not enough arguments to -dev");
		return 0;
	    }
	    device=argv[++i];
	    continue;
	}
	printf("Unknown option %s\n", argv[i]);
	return 0;
    }
	
    foo=popen(winname, "r");
    fgets(string, 255, foo);
    memset(string, 0, sizeof(string));
    fgets(string, 22, foo);
    fscanf(foo,"%x",&thewindow);
    pclose(foo);
    if(thewindow==0) 
    {
	puts("Error!  Can't find window to send events to!");
	return 0;
    }

    if((jsfd=open(device,O_RDONLY))==-1)
    {
	printf("Error opening %s!\n", device);
	puts("Are you sure you have joystick support in your kernel?");
	return 0;
    }
    memset(&js0, 0, sizeof(struct JS_DATA_TYPE));
    memset(&event, 0, sizeof(event));
    event.xkey.type=KeyPress;
    event.xkey.root=RootWindow(thedisp, thescreen);
    event.xkey.window=thewindow;
    event.xkey.subwindow=thewindow;
    event.xkey.same_screen=True;
    event.xkey.time=CurrentTime;

    arup=XKeysymToKeycode(thedisp, arup);
    ardown=XKeysymToKeycode(thedisp, ardown);
    arleft=XKeysymToKeycode(thedisp, arleft);
    arright=XKeysymToKeycode(thedisp, arright);
    but1=XKeysymToKeycode(thedisp, but1);
    but2=XKeysymToKeycode(thedisp, but2); 
#ifdef USE_FOUR_BUTTONS
    but3=XKeysymToKeycode(thedisp, but3); /* craiger */
    but4=XKeysymToKeycode(thedisp, but4);
#endif
    waittime=(long)1000000/waittime;
    autorepeattime=(long)1000000/autorepeattime;

    puts("Initialization complete, entering main loop, ^C to exit...");

    gettimeofday(&timeval[0], &timezone);
    lasttime=&timeval[0];
    timeindex=1;
    presenttime=&timeval[1];
    signal(SIGINT, cleanup);
    signal(SIGTERM, cleanup);
    
    for(;;)
    {
	oldx=js0.x;
	oldy=js0.y;
	oldb=js0.buttons;
	oldb1=b1;
	oldb2=b2;
	oldb3=b3;
	oldb4=b4;
	gettimeofday(presenttime, &timezone);

	elapsed=(((presenttime->tv_sec -lasttime->tv_sec) * 1000000 +
		  (presenttime->tv_usec -lasttime->tv_usec)) > autorepeattime);

	read(jsfd, &js0, JS_RETURN);
	/* craiger mods, use 4 buttons, read simultaneous pressure */
	b1 = (js0.buttons & 0x01);
	b2 = (js0.buttons & 0x02);
#ifdef USE_FOUR_BUTTONS
	b3 = (js0.buttons & 0x04);
	b4 = (js0.buttons & 0x08);
#endif
	if(js0.x>xthreshhi && (oldx<xthreshhi || (autorepeatflag && justafterelapsed)))
	{
	    event.xkey.type=KeyPress;
	    event.xkey.keycode=arright;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyPressMask;
	    XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
	}
	if(oldx>xthreshhi && (js0.x<xthreshhi || (autorepeatflag && elapsed)))
	{
	    event.xkey.type=KeyRelease;
	    event.xkey.keycode=arright;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyReleaseMask;
	    XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
	}
	if(js0.x<xthreshlow && (oldx>xthreshlow || (autorepeatflag && justafterelapsed)))
	{
	    event.xkey.type=KeyPress;
	    event.xkey.keycode=arleft;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyPressMask;
	    XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
	}
	if(oldx<xthreshlow && (js0.x>xthreshlow || (autorepeatflag && elapsed)))
	{
	    event.xkey.type=KeyRelease;
	    event.xkey.keycode=arleft;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyReleaseMask;
	    XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
	}
	if(js0.y>ythreshhi && (oldy<ythreshhi || (autorepeatflag && justafterelapsed)))
	{
	    event.xkey.type=KeyPress;
	    event.xkey.keycode=ardown;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyPressMask;
	    XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
	}
	if(oldy>ythreshhi && (js0.y<ythreshhi || (autorepeatflag && elapsed)))
	{
	    event.xkey.type=KeyRelease;
	    event.xkey.keycode=ardown;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyReleaseMask;
	    XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
	}
	if(js0.y<ythreshlow && (oldy>ythreshlow || (autorepeatflag && justafterelapsed)))
	{
	    event.xkey.type=KeyPress;
	    event.xkey.keycode=arup;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyPressMask;
	    XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
	}
	if(oldy<ythreshlow && (js0.y>ythreshlow || (autorepeatflag && elapsed)))
	{
	    event.xkey.type=KeyRelease;
	    event.xkey.keycode=arup;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyReleaseMask;
	    XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
	}
	if(b1  && (!oldb1 || (autorepeatflag && justafterelapsed)))
	{
	    event.xkey.type=KeyPress;
	    event.xkey.keycode=but1;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyPressMask;
	    XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
	}
	if(oldb1 && (!b1 || (autorepeatflag && elapsed)))
	{
	    event.xkey.type=KeyRelease;
	    event.xkey.keycode=but1;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyReleaseMask;
	    XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
	}
	if(b2 && (!oldb2 || (autorepeatflag && justafterelapsed)))                     
	{
	    event.xkey.type=KeyPress;
	    event.xkey.keycode=but2;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyPressMask;
	    XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
	}
	if(oldb2 && (!b2 || (autorepeatflag && elapsed)))
	{
	    event.xkey.type=KeyRelease;
	    event.xkey.keycode=but2;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyReleaseMask;
	    XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
	}
#ifdef USE_FOUR_BUTTONS
	if(b3 && (!oldb3 || (autorepeatflag && justafterelapsed)))                     
	{
	    event.xkey.type=KeyPress;
	    event.xkey.keycode=but3;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyPressMask;
	    XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
	}
	if(oldb3 && (!b3 || (autorepeatflag && elapsed)))
	{
	    event.xkey.type=KeyRelease;
	    event.xkey.keycode=but3;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyReleaseMask;
	    XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
	}
	if(b4 && (!oldb4 || (autorepeatflag && justafterelapsed)))                     
	{
	    event.xkey.type=KeyPress;
	    event.xkey.keycode=but4;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyPressMask;
	    XSendEvent(thedisp, thewindow, False, KeyPressMask, &event);
	}
	if(oldb4 && (!b4 || (autorepeatflag && elapsed)))
	{
	    event.xkey.type=KeyRelease;
	    event.xkey.keycode=but4;
	    event.xkey.display=thedisp;
	    event.xkey.state=KeyReleaseMask;
	    XSendEvent(thedisp, thewindow, False, KeyReleaseMask, &event);
	}
#endif

/* craiger */
/*                printf("Buttons: %d, %d, %d, %d.\n",b1,b2,b3,b4); */


	XFlush(thedisp);
	if(elapsed) {
	    lasttime=presenttime;
	    timeindex^=1;
	    presenttime=&timeval[timeindex];
	    justafterelapsed=TRUE;
	}
	else
	{
	    justafterelapsed=FALSE;
	}
	
	usleep(waittime);
    }
}

void cleanup(int s)
{
    printf("\n%s caught, cleaning up & quitting.\n", 
	   s==SIGINT ? "SIGINT" : (s==SIGTERM ? "SIGTERM" : "Unknown"));
    XFlush(thedisp);
    XCloseDisplay(thedisp);
    exit(0);
}
